#include "calc.h"
#include "gammatable.h"

/* extern float gammatable[10001][2]; */

extern char bases[]; /* ACGUTRYMKWSBDHVNacgutrymkwsbdhvnt */
extern int gotoh_Table[128][128];

/*
  A  C  G  U  T  R  Y  M  K  W  S  B  D  H  V  N  
 65 67 71 85 84 82 89 77 75 87 83 66 68 72 86 78
  a  c  g   u   t   r   y   m   k   w   s   b   d   h   v   n   t
 97 99 103 117 116 114 121 109 107 119 115 98  100 104 118 110 116
*/

/* conservative statistical estimation of score distribution */

int nalns = 0;

/* -------------------------------------------------------------------------- */

void gotoh_initMatchTable(gotoh_settings gs) {

  int i, j, l, ni, nj, score;
  char *m=NULL;

  for(i=0; i<128; i++) {
    for(j=0; j<128; j++) {
      if(i>j) continue;
      gotoh_Table[j][i] = gs.mismatch;
      gotoh_Table[i][j] = gs.mismatch;
    }
  }

  for(i=0; i<16; i++) {
    ni = bases[i]; /* i pos in table */
    for(j=0; j<16; j++) {
      if(i>j) continue;
      nj = bases[j]; /* j pos in table */
      if(i < 5 && i == j) {
	gotoh_Table[ni][nj] = gs.match;
	gotoh_Table[ni+32][nj] = gs.match;
	gotoh_Table[ni][nj+32] = gs.match;
	gotoh_Table[nj][ni] = gs.match;
	gotoh_Table[nj+32][ni] = gs.match;
	gotoh_Table[nj][ni+32] = gs.match;
	gotoh_Table[ni+32][nj+32] = gs.match;
	gotoh_Table[nj+32][ni+32] = gs.match;
      } else if(i < 11 && i == j) {
	score = floor(0.5 * gs.match + 0.5 * gs.mismatch);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      } else if(i < 15 && i == j) {
	score = floor((float) gs.match / 3 + (float)gs.mismatch / 3);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      } else if(i == 16 && j == 16) {
	score = floor((float) gs.match / 4 + (float)gs.mismatch / 4);
	gotoh_Table[ni][nj] = score;
	gotoh_Table[ni+32][nj] = score;
	gotoh_Table[ni][nj+32] = score;
	gotoh_Table[nj][ni] = score;
	gotoh_Table[nj+32][ni] = score;
	gotoh_Table[nj][ni+32] = score;
	gotoh_Table[ni+32][nj+32] = score;
	gotoh_Table[nj+32][ni+32] = score;
      }
    }
  }
  
  /* [U,T] -> [u,t] = match ! */
  gotoh_Table[84][85] = gs.match;
  gotoh_Table[84][117] = gs.match;
  gotoh_Table[85][84] = gs.match;
  gotoh_Table[85][116] = gs.match;
  gotoh_Table[116][85] = gs.match;
  gotoh_Table[117][84] = gs.match;
  gotoh_Table[116][117] = gs.match;
  gotoh_Table[117][116] = gs.match;
  /* [u,t] -> [U,T]  = match ! */
  
  
  /* for(i=0; i<16; i++) { */
/*     for(j=0; j<16; j++) { /\* go through complete array *\/ */
/*       if(i>j) continue; */
/*       if(i<5 && i == j) { */
/* 	/\* bases[i]; *\/ */
/* 	ni = toascii(bases[i]); */
/* 	nj = toascii(bases[j]); */
/* 	gotoh_Table[i][j] = gs.match; */
/* 	gotoh_Table[i+16][j] = gs.match; */
/* 	gotoh_Table[i][j+16] = gs.match; */
/* 	gotoh_Table[j][i] = gs.match; */
/* 	gotoh_Table[j+16][i] = gs.match; */
/* 	gotoh_Table[j][i+16] = gs.match; */
/* 	gotoh_Table[i+16][j+16] = gs.match; */
/* 	gotoh_Table[j+16][i+16] = gs.match; */
/*       } else { */
/* 	gotoh_Table[i][j] = gs.mismatch; */
/* 	gotoh_Table[i+16][j] = gs.mismatch; */
/* 	gotoh_Table[i][j+16] = gs.mismatch; */
/* 	gotoh_Table[j][i] = gs.mismatch; */
/* 	gotoh_Table[j+16][i] = gs.mismatch; */
/* 	gotoh_Table[j][i+16] = gs.mismatch; */
/* 	gotoh_Table[i+16][j+16] = gs.mismatch; */
/* 	gotoh_Table[j+16][i+16] = gs.mismatch; */
/*       } */
/*     } */
/*   } */

  /* [U,T] -> [u,t] = match ! */
  /* gotoh_Table[3][4] = gs.match; */
/*   gotoh_Table[4][3] = gs.match; */
/*   gotoh_Table[19][4] = gs.match; */
/*   gotoh_Table[20][3] = gs.match; */
/*   /\* [u,t] -> [U,T]  = match ! *\/ */
/*   gotoh_Table[3][20] = gs.match; */
/*   gotoh_Table[4][19] = gs.match; */
/*   gotoh_Table[19][20] = gs.match; */
/*   gotoh_Table[20][19] = gs.match; */

/*   printf("       "); */
/*   for(i=64; i<128; i++) */
/*     printf("%2d:%c\t",i,i); */
/*   printf("\n"); */
/*   for(i=64; i<128; i++) { */
/*     printf("%2d:%c   ",i,i); */
/*     for(j=64; j<128; j++) { */
/*       printf("%3d\t",gotoh_Table[i][j]); */
/*     } */
/*     printf("\n"); */
/*   } */
  
/*   exit(0); */
  return;

}

/* -------------------------------------------------------------------------- */

void gotoh_scanDbase(char * db_seq, int n, char * q_seq, int m,
		     char *db_name, char *q_name, int * score_distr,
		     int n_s, gotoh_settings gs, gotoh_aln List[], int n2add, int Len) {

  int *fw_scores_db=NULL, *bw_scores_db=NULL,mH,i;
  char * rev_q_seq=NULL;
  
  mH = (int) floor((float)m/2) + 1;

  fw_scores_db = gotoh_fwRecursion(db_seq, n, q_seq, m, db_name, q_name, score_distr, n_s, gs);
  
  /* for(i=1;i<=n;i++) */
/*     printf("FW: %d %d\n",i,fw_scores_db[i]); */
/*   printf("&\n"); */

  /* for(i=0; i<MAX_ALNS; i++) { */
/*     if(List[i].alnscore == -1) break; */
/*     gotoh_printGotohAln(List[i], 1); */
/*   } */
  
  gotoh_findLocalMaxima(fw_scores_db, List, n2add, 0, n, db_seq, db_name, q_seq, q_name, m, '+', gs, Len);

  /* check other strand, seq often shorter than db -> reverse query, not db! */
  rev_q_seq = goto_revCompl(q_seq, gs.verbose, m);

  bw_scores_db = gotoh_fwRecursion(db_seq, n, rev_q_seq, m, db_name, q_name, score_distr, n_s, gs);


  gotoh_findLocalMaxima(bw_scores_db, List, n2add, 0, n, db_seq, db_name, rev_q_seq, q_name, m, '-', gs, Len);

  /* for(i=1;i<=n;i++) */
/*     printf("BW: %d %d\n",i,bw_scores_db[i]); */

 /*  for(i=0; i<MAX_ALNS; i++) { */
/*     if(List[i].alnscore == -1) break; */
/*     gotoh_printGotohAln(List[i], 1); */
/*   } */
/*   return; */
  /* fprintf(stderr, "NUMBEROFALNS: %d\n",nalns); */
  /* free temporary arrays */
  free(bw_scores_db);
  free(fw_scores_db);
  free(rev_q_seq);

}

/* -------------------------------------------------------------------------- */

int * gotoh_getMaxS(int scores[], int s, int e) {

  int i, *ret=NULL;

  ret = (int *) calloc(2,sizeof(int));
  ret[0] = 0;  /* position in database */
  ret[1] = -1; /* score at position ret[0] */
  
  for(i=s; i<=e; i++)
    if(scores[i] >= ret[1]) { /* get last max (if equal) */
      ret[0] = i; ret[1] = scores[i];
    }

  return ret;  
  
}

/* -------------------------------------------------------------------------- */

int * gotoh_fwRecursion(char * s, int n, char * q, int m, char *db_name, char *q_name,
			int * score_distr, int n_s, gotoh_settings cfg) {

  int  i,j,p,pp,bi,bj;
  int *S, *S_, *D, *D_, *F, *F_, *tmp;
  int *score, transf_score, minAS;

  S   = (int *) calloc(m+1, sizeof(int));
  S_  = (int *) calloc(m+1, sizeof(int));
  D   = (int *) calloc(m+1, sizeof(int));
  D_  = (int *) calloc(m+1, sizeof(int));
  F   = (int *) calloc(m+1, sizeof(int));
  F_  = (int *) calloc(m+1, sizeof(int));

  score = (int *) calloc(n+1, sizeof(int));
  for(i=0; i<=n; i++) score[i] = -1;

  minAS = abs( 2 * (cfg.gap_open + ((m - 1) * cfg.gap_extend)));
  
  /* initialize semi-local alignment */
  D_[0] = -(cfg.inf);
  F_[0] = -(cfg.inf);
  S_[0] = 0;

  for(j=1;j<=m;j++) { 
    D_[j] = -(cfg.inf);
    F_[j] = (j-1)*cfg.gap_extend + cfg.gap_open;
    S_[j] = F[j];
  }

  for(i=1;i<=n;i++) {
    /* init */
    D[0] = 0;  /* local alignment!*/
    F[0] = -cfg.inf;
    S[0] = 0;

    /* align query  against each database position i */
    for(j=1;j<=m;j++) {
      /* int p,pp; */
      p  = S[j-1] + cfg.gap_open;
      pp = D[j-1] + cfg.gap_extend;
      if(p>pp) D[j] = p;   /* start new insertion */
      else D[j] = pp;      /* extend insertion    */
      
      p = S_[j]  + cfg.gap_open;
      pp = F_[j]  + cfg.gap_extend;
      if(p>pp) F[j] = p;   /* start new deletion */
      else F[j] = pp;      /* extend deletion    */
      
      p = S_[j-1];
      /* if(s[i-1]==q[j-1] && s[i-1] != 'N') p += cfg.match;  /\* match     *\/ */
/*       else p += gotoh_Table[s[i-1]][q[j-1]]; */
      /* /\* else p += cfg.mismatch;  *\/  */           /* mismatch  */
      /* get match/mismatch score from lookuptable */
      p += gotoh_Table[s[i-1]][q[j-1]];
      
      if(D[j]>p) p=D[j];                   /* insertion */
      if(F[j]>p) p=F[j];                   /* deletion  */
      
      S[j] = p;
    }
    
    /* transform score */
    transf_score = S[m] + minAS;

    score[i] = transf_score;
    /* nalns++; */
    /* score[i] = S[m]; */
    
    score_distr[transf_score]++; /* sum up to transformed score distribution */

    /* rotate memory */
    tmp = S_; S_ = S; S = tmp;
    tmp = D_; D_ = D; D = tmp;
    tmp = F_; F_ = F; F = tmp;
    
  }

 /*  for(i=1;i<=n;i++) printf("%d: %d\n",i,score[i]); */

  free(S);free(S_);free(D);free(D_);free(F);free(F_);
  
  return score;
  
}

/* -------------------------------------------------------------------------- */

float * gotoh_bwRecursion(char * s, char * q, int m, int s_pos, gotoh_settings cfg,
			  char *ss, char *qq, int sollscore, int n2add) {

  int i,j,k,len, p, pp, wc;
  int **S, **D, **F;
  char *sa, *qa, /* *sx, */ btstate;
  int si,qj,found, start,gos,mms,perc;
  float *res;
  int db_end;

  len = 2 * m + 1;
  if(len > s_pos) len = s_pos;

  /* retransform score */
  wc = abs(2*(cfg.gap_open + (m-1)*cfg.gap_extend));
  sollscore -= wc;

  S = (int **) calloc(len+2, sizeof(int *));
  D = (int **) calloc(len+2, sizeof(int *));
  F = (int **) calloc(len+2, sizeof(int *));
  for(i=0;i<=len;i++) {
    S[i] = (int *) calloc(m+2, sizeof(int *));
    D[i] = (int *) calloc(m+2, sizeof(int *));
    F[i] = (int *) calloc(m+2, sizeof(int *));
  }
  sa = (char *) calloc(2*m+len+2, sizeof(int *));
  qa = (char *) calloc(2*m+len+2, sizeof(int *));
  
  D[0][0] = (-cfg.inf);
  F[0][0] = (-cfg.inf);
  S[0][0] = 0;

  for(j=1;j<=m;j++) { 
    D[0][j] = (-cfg.inf);
    F[0][j] = (j-1)*cfg.gap_extend + cfg.gap_open;
    S[0][j] = F[0][j];
  }

  for(i=1;i<=len;i++) {
    si = s_pos-(i-1);
    
    D[i][0] = 0;
    F[i][0] = (-cfg.inf);
    S[i][0] = 0;
    
    for(j=1;j<=m;j++) {
      qj = m-j;	 

      p  = S[i][j-1] + cfg.gap_open;
      pp = D[i][j-1] + cfg.gap_extend;
      D[i][j] = p > pp ? p : pp;
      
      p  = S[i-1][j] + cfg.gap_open;
      pp = F[i-1][j] + cfg.gap_extend;
      F[i][j] = p > pp ? p : pp;

     /*  if(s[si] == q[qj] && s[si] != 'N') p = cfg.match; */
/*       else p = gotoh_Table[s[si]][q[qj]]; */
      p = gotoh_Table[s[si]][q[qj]];
      /* p = s[si] == q[qj] ? cfg.match : cfg.mismatch; */
      /* if(s[si] == q[qj] && s[si] != 'N') p = cfg.match; */
/*       else p = cfg.mismatch; */
      p += S[i-1][j-1];

      
      if(D[i][j]>p) p=D[i][j];
      
      if(F[i][j]>p) p=F[i][j];
      S[i][j] = p;
    }
			 
  }
  
  i=0; found = 0;
  while(i<len) {
    if(S[i][m] == sollscore) {
      found = 1;
      break;
    }
    i++;
  }
  /* i is now the position with our requested score !!! */
  if(found == 0 && cfg.verbose == 1) {
    fprintf(stderr, "WARNING: At position i=%d, there is a better alignment possible\n",s_pos);
    fprintf(stderr, "         than in forward recursion calculated\n");
  }

  start = s_pos-i+1;
  db_end = start;
  
  res = (float *) calloc(7, sizeof(float));
  /* percId, mm, go, length */
  gos = 0;
  mms = 0;
  perc = 0;
  
  j = m;
  k = 0;
  btstate = 'S';


  while( (j>0) ) {

    if(i<1) break;
    
    si = s_pos -(i-1);
    qj = m -j;

    if(si < 0) {
      while(j>0) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
	j--;
	k++;
	qj = m -j;
	si = 1;
      }
      break;
    }

    switch(btstate) {
    case 'S':
      /* if( S[i][j] == S[i-1][j-1] + cfg.match && s[si] == q[qj] && s[si] != 'N') { */
/* 	sa[k] = s[si]; */
/* 	qa[k] = q[qj]; */
/* 	db_end++; */
/* 	i--; */
/* 	j--; */
/* 	k++; */
/* 	btstate = 'S'; */
/* 	perc++; */
/*       } else */ if(S[i][j] == S[i-1][j-1] + gotoh_Table[s[si]][q[qj]]) {
	sa[k] = s[si];
	qa[k] = q[qj];
	db_end++;
	i--;
	j--;
	k++;
	btstate = 'S';
	if(gotoh_Table[s[si]][q[qj]] <= gotoh_Table[78][78])
	  mms++;
	else
	  perc++;
      }
/*       else if( S[i][j] == S[i-1][j-1] + cfg.mismatch && s[si] != q[qj]) { */
/* 	sa[k] = s[si]; */
/* 	qa[k] = q[qj]; */
/* 	db_end++; */
/* 	i--; */
/* 	j--; */
/* 	k++; */
/* 	btstate = 'S'; */
/* 	mms++; */
/*       } */
      /* if(S[i][j] == S[i-1][j-1] + gotoh_Table[s[si]][q[qj]]) { */
/* 	sa[k] = s[si]; */
/* 	qa[k] = q[qj]; */
/* 	db_end++; */
/* 	i--; */
/* 	j--; */
/* 	k++; */
/* 	btstate = 'S'; */
/* 	if(gotoh_Table[s[si]][q[qj]] <= gotoh_Table[78][78]) */
/* 	  mms++; */
/* 	else */
/* 	  perc++; */
/*       } */
      else if( S[i][j] == F[i][j] ) {
	btstate = 'F';
      }
      else if( S[i][j] == D[i][j] ) {
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in S[%d][%d]=%d\n",i,j,S[i][j]);
	exit(0);
      } 
      break;
    case 'D':
      if(D[i][j] ==  S[i][j-1] + cfg.gap_open ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
        j--;	
	k++;
	btstate = 'S';
	gos++;
      }
      else if(D[i][j] == D[i][j-1] + cfg.gap_extend ) {
	sa[k] = '-';
	qa[k] = q[qj];
	start++;
	j--;
	k++;
	btstate = 'D';
      }
      else {
	fprintf(stderr,"This should not happen in D[%d][%d]=%d\n",i,j,D[i][j]);
	exit(0);
      } 
      break;
    case 'F':
      if(F[i][j] ==  S[i-1][j]   + cfg.gap_open ) {
	sa[k] = s[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'S';
	gos++;
      }
      else if ( F[i][j] == F[i-1][j]  + cfg.gap_extend ) {
	sa[k] = s[si];
	qa[k] = '-';
	db_end++;
	i--;
	k++;
	btstate = 'F';
      }
      else {
	fprintf(stderr,"This should not happen in F[%d][%d]=%d\n",i,j,F[i][j]);
	exit(0);
      }    
      break;
    default:
      fprintf(stderr,"This should not happen at all. btstate: %c\n",btstate);
      exit(0);
    }
    
  }

  k = strlen(sa);

  res[0] = (float) perc / (float) m * 100;
  res[1] = (float) mms;
  res[2] = (float) gos;
  res[3] = (float) k;
  res[4] = (float) start + 1;
  res[5] = (float) db_end;

  strcpy(ss, sa);
  strcpy(qq, qa);

  for(i=0;i<=len+1;i++) {
    free(S[i]);
    free(D[i]);
    free(F[i]);
  }
  free(S);free(D);free(F);
  free(sa);free(qa);

  return res;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_findLocalMaxima(int *scores, gotoh_aln List[], int n2add,
			   int s, int e,
			   char *db_seq, char *db_name,
			   char *q_seq, char *q_name, int m,
			   char strand, gotoh_settings gs, int Len) {

  int ms, mi, nsl, nel, nsr, ner, ii;
  
  gotoh_getMaxLinear(scores, s, e, &ms, &mi); /* insert/backtrace ms an mi */

  ii=0; /* find place in L to put local maxima */
  while(List[ii].alnscore >= ms && ii < MAX_ALNS) ii++;
  /* try to prevent including duplicated entries !!*/
  if(ii < MAX_ALNS) { /* insert */
      
    gotoh_backtraceAln2(ii, mi, ms, strand, db_name, q_name, List, m, db_seq, q_seq, gs, n2add, Len);
    
  } else { /* stop inserting if list full with higher scores! */
    
    return;
      
  }
  
  /* d&c recursion on left side */
  nsl = s; nel = mi - 2*m;
  if(nel - nsl >= m) 
    gotoh_findLocalMaxima(scores, List, n2add, nsl, nel, db_seq, db_name, q_seq, q_name, m, strand, gs, Len);
  
  /* d&c recursion on right side */
  nsr = mi + 2*m + 1; ner = e;
  if(ner - nsr >= m) 
    gotoh_findLocalMaxima(scores, List, n2add, nsr, ner, db_seq, db_name, q_seq, q_name, m, strand, gs, Len);
  
  return;

}

/* -------------------------------------------------------------------------- */

void gotoh_getMaxLinear(int *scores, int s, int e, int *ms, int *mi) {

  int i, sm=0, im=-1;

  for(i=s; i<=e; i++) {
    if(scores[i] >= sm) {
      sm = scores[i];
      im = i;
    }
  }

  *ms = sm; *mi = im;
  
  return;
  
}

/* -------------------------------------------------------------------------- */
/**
   q_seq is sense if strand eq '+', otherwise it is already in antisense 
 */
void gotoh_backtraceAln2(int ii, int imax, int smax, char strand, char *s_name, char* q_name,
			 gotoh_aln List[], int m, char *db_seq, char *q_seq,
			 gotoh_settings gs, int n2add, int Len) {

  float *bt;
  gotoh_aln tmpaln, curaln;
  char *revseq1=NULL,*revseq2=NULL;
  int ll;

  curaln = gotoh_initL();

  curaln.alnscore = smax;
  curaln.db_len = Len;
  /* curaln.sEnd = n2add+imax; */
  curaln.qStart = 1;
  curaln.qEnd = m;
  curaln.strand = strand;
  curaln.hitname = (char *) calloc(strlen(s_name)+1, sizeof(char));
  strcpy(curaln.hitname, s_name);
  curaln.queryname = (char *) calloc(strlen(q_name)+1, sizeof(char));
  strcpy(curaln.queryname, q_name);
  curaln.hitseq = (char *) calloc(2*m+1, sizeof(char));
  curaln.queryseq = (char *) calloc(2*m+1, sizeof(char));
  bt = gotoh_bwRecursion(db_seq, q_seq, m, imax-1, gs, curaln.hitseq, curaln.queryseq, smax, n2add);
  if(strand == '-') {
    ll = strlen(curaln.hitseq);
    revseq1 = goto_revCompl(curaln.hitseq, gs.verbose, ll);
    revseq2 = goto_revCompl(curaln.queryseq, gs.verbose, ll);
    strcpy(curaln.hitseq,revseq1);
    strcpy(curaln.queryseq,revseq2);
    
    free(revseq1); free(revseq2);
  }
  curaln.percID = bt[0]; curaln.mismatches = (int) bt[1];
  curaln.gapopening = (int) bt[2]; curaln.length = (int) bt[3];
  curaln.sStart = n2add + (int) bt[4] - 1; 
  curaln.sEnd = n2add + (int) bt[5];
  curaln.evalue = 999999;
  free(bt);

  /* don't insert duplicated entry (due to splits system!) */
  if(ii>=1 && List[ii-1].alnscore == curaln.alnscore && List[ii-1].sEnd == curaln.sEnd)
    return;

  if(ii == 0 || List[ii].alnscore == -1) { /* just insert and goto next db position for next aln */
    List[ii] = curaln;
    return;
  }

  tmpaln = gotoh_initL();

  tmpaln = List[ii];
  List[ii] = curaln;
  ii++;
  while(ii < MAX_ALNS) {
    if(List[ii].alnscore == -1) {
      List[ii] = tmpaln;
      return;
    }
    curaln = gotoh_initL();
    curaln = List[ii];
    
    List[ii] = tmpaln;

    tmpaln = gotoh_initL();
    tmpaln = curaln;

    ii++;
  }

  free(tmpaln.hitname); free(tmpaln.queryname); free(tmpaln.hitseq); free(tmpaln.queryseq);

  return;
  
}

/* -------------------------------------------------------------------------- */

char * goto_revCompl(char *sequence, int v, int len) {

  int i,j;
  char nt;
  char *result;

  result = (char*) calloc(len+1, sizeof(char));
  
  for(i=len-1,j=0;i>=0;i--,j++) {
    switch(sequence[i]) {
    case 'a' : nt = 'u' ; break;
    case 'A' : nt = 'U' ; break;
    case 'u' : nt = 'a' ; break;
    case 'U' : nt = 'A' ; break;
    case 'c' : nt = 'g' ; break;
    case 'C' : nt = 'G' ; break;
    case 'g' : nt = 'c' ; break;
    case 'G' : nt = 'C' ; break;
    case '-' : nt = '-' ; break;
    default :
      gotoh_Warning("There are strange letters in your sequence. Parsed to 'N'.", v);
      nt='N' ;
    }
    result[j]=nt;
  }

  return result;
  
}

/* -------------------------------------------------------------------------- */
/* score_distr => data, distribution of all possible scores */
double * gotoh_getParams(int * score_distr, int n_s, int n_q, gotoh_settings gs, char * qname) {

  int minscore,maxscore, i, mins, maxs, minAS, maxN=0,maxn;
  double eps = 0.000001;
  double subs=0;
  double t;
  double *params, k, phi, phiprime, gamma;
  char *datafile=NULL;
  FILE *fp_data=NULL;
  gotoh_fitData fd;


   
  /* truncate score_distr to get only data for fitting */
  /* use [0.55*n_s, 0.85*n_s] as limits */
  /* minscore = (int) floor(0.605 * n_s); */
/*   maxscore = (int) floor(0.85 * n_s); */
  /*  maxscore = (int) floor(0.95 * n_s); */
  /* calculate transformation for retransforming score */
  minAS = abs(2 * (gs.gap_open + (n_q - 1) * gs.gap_extend));

  /* get start and end where 90% of the data is located
     centered at the max */
  gotoh_getPercRange(score_distr, n_s, &minscore, &maxscore, minAS, qname, &maxN, &maxs);

  fprintf(stderr, "MIN: %d MAX: %d size: %d minAS: %d\n",minscore, maxscore,n_s, minAS);
 
  /* for(i=minscore; i<=maxscore; i++) { */
/*     if(score_distr[i] >= maxN) */
/*       maxN = score_distr[i]; */
/*     if(score_distr[i] > 0) */
/*       maxs = i; */
/*   } */

  maxs += 20 - minAS;
  
  if(gs.output_data == 1) {
    mins = floor((float)(minscore - minAS) / 10) * 10;
    /* maxs = floor((float)(-minAS + n_s)/10) * 10 + 20; */
    maxn = floor(log(maxN)) + 2;
    datafile = (char *) calloc(strlen(qname)+16, sizeof(char));
    strcpy(datafile,qname);
    strcat(datafile, ".GotohScan.agr");
    fp_data = fopen(datafile, "w");
    fprintf(fp_data, "# Grace project file\n");
    fprintf(fp_data, "#\n");
    fprintf(fp_data, "@version 50120\n");
    fprintf(fp_data, "@page size 792, 612\n");
    fprintf(fp_data, "@map font 4 to \"Helvetica\", \"Helvetica\"\n");
    fprintf(fp_data, "@map font 5 to \"Helvetica-Oblique\", \"Helvetica-Oblique\"\n");
    fprintf(fp_data, "@with g0\n");
    fprintf(fp_data, "@ world %d, -3, %d, %d\n",mins,maxs,maxn);
    fprintf(fp_data, "@ stack world 0, 0, 0, 0\n");
    fprintf(fp_data, "@ znorm 1\n");
    fprintf(fp_data, "@ view 0.150000, 0.150000, 1.150000, 0.850000\n");
    fprintf(fp_data, "@ title \"GotohScan - Evalue calculation\"\n");
    fprintf(fp_data, "@ title font 4\n");
    fprintf(fp_data, "@ subtitle \"%s\"\n",qname);
    fprintf(fp_data, "@ subtitle font 5\n");
    fprintf(fp_data, "@ yaxes scale Normal\n");
    fprintf(fp_data, "@ xaxes scale Normal\n");
    fprintf(fp_data, "@ xaxis label \"alignment score\"\n");
    fprintf(fp_data, "@ xaxis label font 4\n");
    fprintf(fp_data, "@ xaxis ticklabel font 4\n");
    fprintf(fp_data, "@ xaxis tick place both\n");
    fprintf(fp_data, "@ xaxis tick major 20\n");
/*     fprintf(fp_data, "@ xaxis tick minor ticks 1\n"); */
    fprintf(fp_data, "@ yaxis label \"log( # alns )\"\n");
    fprintf(fp_data, "@ yaxis label font 4\n");
    fprintf(fp_data, "@ yaxis ticklabel font 4\n");
    fprintf(fp_data, "@ yaxis tick place both\n");
    fprintf(fp_data, "@ yaxis tick major 10\n");
/*     fprintf(fp_data, "@ yaxis tick minor ticks 1\n"); */
    fprintf(fp_data, "@ legend on\n");
    fprintf(fp_data, "@ legend font 4\n");
    fprintf(fp_data, "@ legend char size 0.9\n");
    fprintf(fp_data, "@ s0 symbol 10\n");
    fprintf(fp_data, "@ s0 symbol size 0.160000\n");
    fprintf(fp_data, "@ s0 line type 0\n");
    fprintf(fp_data, "@ s0 legend \"Score distribution (Max=%d)\"\n",n_s-minAS);
    fprintf(fp_data, "@ s1 line linewidth 1.5\n");
    fprintf(fp_data, "@ s1 line color 2\n");
    fprintf(fp_data, "@ s1 line style 2\n");
    fprintf(fp_data, "@ s1 legend \"Fitted gamma distribution\"\n");

    fprintf(fp_data, "@ s2 line type 1\n");
    fprintf(fp_data, "@ s2 line style 3\n");
    fprintf(fp_data, "@ s2 line linewidth 1.0\n");
    fprintf(fp_data, "@ s2 line color 7\n");
    fprintf(fp_data, "@ s2 line pattern 1\n");
    fprintf(fp_data, "@ s2 legend \"Portion used for fitting\"\n");

    fprintf(fp_data, "@ s3 line type 1\n");
    fprintf(fp_data, "@ s3 line style 3\n");
    fprintf(fp_data, "@ s3 line linewidth 1.0\n");
    fprintf(fp_data, "@ s3 line color 7\n");
    fprintf(fp_data, "@ s3 line pattern 1\n");

    fprintf(fp_data, "@target G0.S0\n");
    fprintf(fp_data, "@type xy\n");
    /* for(i=minscore;i<=n_s;i++)  */
    for(i=1;i<=n_s;i++) 
      if(score_distr[i] > 0) fprintf(fp_data, "%d %lf\n",i-minAS,log(score_distr[i]));
    fprintf(fp_data, "&\n"); /* seperate data from fitted curve */

  }

  
  fd = gotoh_initFit(score_distr, minscore, maxscore, gs.verbose);
  
  /* estimate k */
  /* start value for k - estimation */
  k = (3. - fd.s + sqrt(fd.s*fd.s + 18.*fd.s + 9.)) / (12. *fd.s);

  if(k == 0)
    gotoh_Error("Value for start k is zero!",gs.verbose);
  
  phi = gotoh_getPhi(k);

  phiprime = gotoh_getPhiPrime(k);

  subs = (log(k) - phi - fd.s) / ( 1/k - phiprime);
  
  while(fabs(subs) > eps) {
    k -= subs;
    phi = gotoh_getPhi(k);
    phiprime = gotoh_getPhiPrime(k);
    subs = (log(k) - phi - fd.s) / ( 1/k - phiprime);
  
  }

  t = 1/k * (double) fd.mean;   /* estimated theta t !!! */
  
  gamma = gotoh_logGammaX(k);

  if(gs.output_data == 1) {
    fprintf(fp_data, "@target G0.S1\n");
    fprintf(fp_data, "@type xy\n");
    /* for(i=minscore;i<=maxscore;i++) */
    for(i=minscore-50;i<=n_s;i++)
      fprintf(fp_data, "%d\t%lf\n",i-minAS,gotoh_getLogGammaVal(i,k,t,gamma,gs.verbose)+log(fd.N));
    fprintf(fp_data, "&\n");

    fflush(fp_data);

    
    fprintf(fp_data, "%d\t%d\n",minscore-minAS,maxs);
    fprintf(fp_data, "%d\t-3\n",minscore-minAS);
    fprintf(fp_data, "&\n");

    fprintf(fp_data, "%d\t%d\n",maxscore-minAS,maxs);
    fprintf(fp_data, "%d\t-3\n",maxscore-minAS);
    fprintf(fp_data, "&\n");

    fflush(fp_data);
    fclose(fp_data);
    fprintf(stderr, "Score distribution and fit data  for query '%s' written to '%s'.\n",qname,datafile);
    free(datafile);
  }

  params = (double *) calloc(4, sizeof(double));

  fprintf(stderr, "theta: %f k: %f\n",t,k);
  
  params[0] = k;
  params[1] = t;
  params[2] = gamma;
  params[3] = fd.N;
  
  return params;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_getPercRange(int *s, int n_s, int * start, int *end, int minAS, char * qname, int *maxC, int *maxI) {

  int i, is, N=0, l, r, sum=0, minPset = 0;
  float curP, minP=0.55, maxP=1, ant=0;
  int pi=-1, mi=0, p, q, ssum, d;
  float d1=0., d2=0., d3=0.;
 
  is = floor(0.55*n_s); /* start at 'is' to avoid including split artefacts! */
  
 
  l = is;
  r = floor(0.85*n_s);

/*   *start = floor(0.66*n_s); */
/*   *end = floor(0.87*n_s); */
/*   return; */

  /* curve is in 5 % of score around maximum */
  /* -> 2.5 to the left and 2.5 to the right */
  l = floor(n_s * 0.02) + 1;
  r = floor(n_s * 0.05) + 1;
  fprintf(stderr,"l: %d r: %d\n",l,r);
  
 /*  printf("is: %d n-s: %d\n",is,n_s); */
  mi = s[n_s-1];
  *maxI = n_s;
  
  for(i=n_s-2; i>=is; i--) {
    if(mi < s[i]) {
      mi = s[i]; pi = i;
    }
    N += s[i];
    if(*maxI == n_s && s[i] > 0) *maxI = i;
  }

  *maxC = mi;
  /* *maxI = pi; */
  
  if(pi - l > 0) *start = pi - l;
  else *start = 0;
  if(pi + r >= n_s) *end = n_s;
  else *end = pi + r;
  
  return;
  /* for(i=1; i<n_s-1; i++) { */
/*      if(s[i] >= 1000) { */
/*        d1 = ((float)(abs(s[i]-s[i-1])+abs(s[i]-s[i+1]))/2); */
/*        d2 = abs(s[i-1] - s[i+1]); /\* 2304 *\/ */
/*        /\* if distance to neighbours higher than 1000*(their distance), */
/* 	  substitute by average of both neighbours *\/ */
/*        	if(d2 == 0 && s[i-1] > 0) */
/* 	  d3 = s[i-1] * 1000; */
/* 	else if(d2 != 0) */
/* 	  d3 = d2 * 1000; */
	
/* 	if(d3 < d1) s[i] = floor( d2/2 ); /\* subtitute artefact by avg of neighbours *\/ */

	
/*      } */
/*      N += s[i]; */
/*      if(s[i] > mi && i>is) { mi = s[i]; pi = i; } */
/*   } */

/*   p = pi; */
/*   q = pi; */
/*   ant = mi / N;  */
/*   while(ant <= 0.25) { */
/*     q++; */
/*     mi += s[q]; */
/*     ant = mi / N; */
/*     if(ant <= 0.25) { */
/*       p--; */
/*       mi += s[p]; */
/*       ant = mi / N; */
/*     }  */
/*   } */

/*   *start = p; */
/*   *end = q; */
  
  {
    char *scorefile=NULL;
    FILE *fp_score;
    /* print score distr */
    scorefile = (char *) calloc(strlen(qname)+32, sizeof(char));
    strcpy(scorefile, qname);
    strcat(scorefile, ".GotohScan.scores.dat");
    fp_score = fopen(scorefile, "w");
    for(i=1; i<=n_s; i++) {
      if(s[i] > 0)
	fprintf(fp_score, "%d %d %d %lf\n", i, i-minAS, s[i], log(s[i]));
      else
	fprintf(fp_score, "%d %d %d -9999999\n", i, i-minAS, s[i]);
    }
    fclose(fp_score);
    free(scorefile);
  }
  
/*   for(i=l; i<n_s; i++) { */
/*     sum += s[i]; */
/*     curP = (float) sum / (float) N; */
/*     /\* if(curP >= minP && minPset == 0) { *\/ */
/* /\*       l = i; *\/ */
/* /\*       minPset = 1; *\/ */
/* /\*     } *\/ */
/*     if(curP >= maxP) { */
/*       r = i; */
/*       break; */
/*     } */
/*   } */

/*   *start = l; */
/*   /\* *end = r; *\/ */
/*   *end = 0.85*n_s; */
    
  return;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_fitData gotoh_initFit(int *sd, int min, int max, int v) {

  gotoh_fitData fd;
  int i;
  float lsum=0;
  
  fd.N = 0.;
  fd.mean = 0.;

  for(i=min; i<=max; i++) {
    fd.N += (float) sd[i];
    fd.mean += (float) i*(float) sd[i];
    lsum += (float) sd[i]*log((float) i);
  }

  if(fd.N == 0 || fd.mean <= 0)
    gotoh_Error("There are no or only zero-valued datapoints available for fitting", v);

  fd.mean /= fd.N;
  fd.s1 = log(fd.mean);
  fd.s2 = lsum / fd.N;
  fd.s = fd.s1 - fd.s2;
  
  return fd;

}

/* -------------------------------------------------------------------------- */

double gotoh_getPhi(double k) {

  if(k >= 8)
    return (log(k) - (1  + (1 - (1/10 - 1/(21*k*k)) / (k*k)) / (6*k)) / (2*k));

  return (gotoh_getPhi(k+1) - 1/k);
  
}

/* -------------------------------------------------------------------------- */
/* for details see: Choi and Wette (1969) */
double gotoh_getPhiPrime(double k) {

  if(k >= 8) 
    return ((1. + (1.+ (1.- (0.2 - 1./(7.*k*k)) / (k*k)) / (3.*k)) / (2.*k)) / k);

  return (gotoh_getPhiPrime(k+1.) + 1./(k*k));
    
}

/* -------------------------------------------------------------------------- */

float gotoh_getGammaVal(int x, double k, float theta, float gamma, int v) {

  float pg=0;
  
  if(k < 0 || theta < 0 || x < 0)
    gotoh_Error("negative value in gamma distribution -> is this possible???", v);

  pg = pow(x, k-1) * exp(-x/theta) / (pow(theta,k) * gamma);
  
  return pg;

}

/* -------------------------------------------------------------------------- */

float gotoh_getLogGammaVal(int x, double k, float theta, float gamma, int v) {

  float pg=0;

  if(k < 0 || theta < 0 || x < 0)
    gotoh_Error("negative value in gamma distribution -> is this possible???",v);

  pg = (k - 1) * log(x) - x/theta - k*log(theta) - gamma;

  return pg;
  
}

/* -------------------------------------------------------------------------- */

double gotoh_gammaX(double x) {

  int i;
  double g=0;
  
  if(x >= 1 && x <= 2) {
    i=0;
    while(1) {
      if(x > gotoh_gammatable[i][0]) { i++; continue; }
      else { break; }
    }
    g = gotoh_gammatable[i][1];
  } else {
    if(x < 1) {
      g = gotoh_gammaX(x+1) / x;
    } else {
      g = (x - 1) * gotoh_gammaX(x - 1);
    }
  }
  
  return g;
  
}
/* -------------------------------------------------------------------------- */

double gotoh_logGammaX(double x) {

  int i;
  double g=0;

  if(x >= 1 && x <= 2) {
    i=0;
    while(1) {
      if(x > gotoh_gammatable[i][0]) { i++; continue; }
      else { break; }
    }
    g = log(gotoh_gammatable[i][1]);
  } else {
    if(x < 1) {
      g = gotoh_logGammaX(x+1) - log(x);
    } else {
      g = log(x - 1) + gotoh_logGammaX(x-1);
    } 
  }
  
  return g;
  
}

/* -------------------------------------------------------------------------- */

double gotoh_uFunction(double a, double z) {

  double u=1., t=1., k=1., eps = 0.000001;

  /* if(z < a) */
/*     fprintf(stderr, "WARNING: estimated parameter k (%lf) < x/theta (%lf)\n. Numerical problems to be expected!\n",a, z); */

  while(t > eps) {
    t *= (a - k) / z;
    u += t;
    k++;
  }

  return u;
  
}

/* -------------------------------------------------------------------------- */

struct gotoh_aln gotoh_initL(void) {

  gotoh_aln ga;

  ga.db_len = 0;
  ga.hitseq=NULL; ga.queryseq=NULL; ga.hitname=NULL; ga.queryname=NULL;
  ga.sStart = -1; ga.sEnd = -1;
  ga.qStart = -1; ga.qEnd = -1;
  ga.length = 0;
  ga.alnscore = -1;
  ga.Levalue = 999999;
  ga.evalue = 1e-3;
  ga.strand = '+';
  ga.mismatches = 0; ga.gapopening = 0;
  ga.percID = 0;
  
  return ga;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_printGotohAln(gotoh_aln aln, int f) {

  int i;
  char *t=NULL;

  switch(f) {
  case 0: /* blast */
    fprintf(stdout, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(stdout, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    if(aln.strand == '+') {
      fprintf(stdout, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    } else {
      fprintf(stdout, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sEnd,aln.sStart);
    }
    fprintf(stdout, "%e\n",aln.evalue);
    break;
  case 1: /* blast + aln sequences */
    fprintf(stdout, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(stdout, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    if(aln.strand == '+') {
      fprintf(stdout, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    } else {
      fprintf(stdout, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sEnd,aln.sStart);
    }
    fprintf(stdout, "%e\t",aln.evalue);
    fprintf(stdout, "strand:%c\t", aln.strand);
    fprintf(stdout, "%s&%s\n",aln.hitseq,aln.queryseq); 
    break;
  case 2: /* fasta */
    fprintf(stdout, ">%s:%s:%d-%d:%c:%.2f:%e\n",aln.queryname,aln.hitname,aln.sStart,aln.sEnd,aln.strand,aln.percID,aln.evalue);
    for(i=0; i<aln.length; i++)
      if(aln.hitseq[i] != '-') fprintf(stdout, "%c",aln.hitseq[i]);
    fprintf(stdout, "\n");
    break;
  case 3: /* maf */
    /* give real sequence lengths and NOT alignment lengths! */
    { int i,ll;
      char *t=NULL;
      t = (char *) calloc(32, sizeof(char));
      strncpy(t, aln.hitname, 30);
      
      fprintf(stdout, "a score=%e\n",aln.evalue);
      fprintf(stdout, "%-30s %13d %5d %c %13d %s\n",t,aln.sStart,aln.length,aln.strand,aln.db_len,aln.hitseq);
      strncpy(t, aln.queryname, 30);
      ll = 0;
      for(i=0; i<aln.length; i++)
	if(aln.queryseq[i] != '-')
	  ll++;
      fprintf(stdout, "%-30s %13d %5d %c %13d %s\n",t,aln.qStart,ll,aln.strand,ll,aln.queryseq);
      free(t);
    }
    fprintf(stdout, "\n");
    /* fprintf(stdout, "CLUSTAL W\n\n\n"); */
/*     t = (char *) calloc(32, sizeof(char)); */
/*     strncpy(t, aln.hitname, 30); */
/*     fprintf(stdout, "%-30s %s\n",t, aln.hitseq); */
/*     strncpy(t, aln.queryname, 30); */
/*     fprintf(stdout, "%-30s %s\n",t, aln.queryseq); */
/*     free(t); */
/*     fprintf(stdout, "%-30s ",""); */
/*     for(i=0;i<aln.length; i++) { */
/*       if(aln.hitseq[i] == aln.queryseq[i]) fprintf(stdout, "*"); */
/*       else fprintf(stdout, " "); */
/*     } */
/*     fprintf(stdout, "\n//\n"); */
    break;
  case 4: /* bed */
    fprintf(stdout, "%s\t%d\t%d\t%s\t%e\t%c\t",aln.hitname,aln.sStart,aln.sEnd,aln.queryname,aln.evalue,aln.strand);
    fprintf(stdout, "note=\"%s&%s\"\n",aln.hitseq,aln.queryseq);
    break;
  case 5: /* gff */
    fprintf(stdout, "%s\tGotohScan\t%s\t%d\t%d\t",aln.hitname,aln.queryname,aln.sStart,aln.sEnd);
    fprintf(stdout, "%e\t%c\t",aln.evalue,aln.strand);
    fprintf(stdout, "note=\"%s&%s\"\n",aln.hitseq,aln.queryseq);
    break;
  case 6: /* my own test output */
    fprintf(stdout, "%s\n%s\n",aln.hitname,aln.hitseq);
    fprintf(stdout, "%s\n%s\n",aln.queryname,aln.queryseq);
    fprintf(stdout, "start: %d length: %d strand: %c ",aln.sStart,aln.length,aln.strand);
    fprintf(stdout, "alnscore: %d log(Evalue): %lf Evalue: %.2e\n\n",aln.alnscore,aln.Levalue,aln.evalue);
    break;
  default: /* blast -> 0 */
    fprintf(stdout, "%s\t%s\t%5.2f\t",aln.queryname,aln.hitname,aln.percID);
    fprintf(stdout, "%d\t%d\t%d\t",aln.length,aln.mismatches,aln.gapopening);
    fprintf(stdout, "%d\t%d\t%d\t%d\t",aln.qStart,aln.qEnd,aln.sStart,aln.sEnd);
    fprintf(stdout, "%e\n",aln.evalue);
  }

  fflush(NULL);
  return;
  
}

/* -------------------------------------------------------------------------- */

void gotoh_getRNA(char *seq, int l) {

  int i,j=0;
  char *tmp=NULL;
  
  i=0; j=0;

  tmp = (char *) calloc(l+1, sizeof(char));
  printf("l: %d\n%s\n",l,seq);
  while(i<l) {
    fprintf(stderr,"%d: %c\n",i,seq[i]);
    switch(seq[i]) {
    case 'a':
      tmp[j] = 'A'; j++;
      break;
    case 'c':
      tmp[j] = 'C'; j++;
      break;
    case 'g':
      tmp[j] = 'G'; j++;
      break;
    case 't':
      tmp[j] = 'U'; j++;
      break;
    case 'u':
      tmp[j] = 'U'; j++;
      break;
    case 'A':
      tmp[j] = 'A'; j++;
      break;
    case 'C':
      tmp[j] = 'C'; j++;
      break;
    case 'G':
      tmp[j] = 'G'; j++;
      break;
    case 'T':
      tmp[j] = 'U'; j++;
      break;
    case 'U':
      tmp[j] = 'U'; j++;
      break;
    case 'N':
      tmp[j] = 'N'; j++;
      break;
    case 'n':
      tmp[j] = 'N'; j++;
      break;
    case ' ':
      break;
    default:
      /* fprintf(stderr, "ERROR: You have strange letters in your sequence\n"); */
/*       fprintf(stderr, "       Letter: '%c' at position: %d\n",seq[i],i); */
/*       exit(0); */
      tmp[j] = 'N'; j++;
    }
    i++;
  }

  tmp[j] = '\0';

  strcpy(seq, tmp);
    
  free(tmp);
  
}

/* -------------------------------------------------------------------------- */
